import argparse
import sys
import time
from datetime import datetime
from typing import Optional, Dict

import requests
from threatray.api import API

from show_report import get_sample_by_analysis

api = API()


def memory_address(val) -> int:
    try:
        if isinstance(val, int):
            return int(val)
        if isinstance(val, str):
            return int(val, 16)
    except ValueError as _:
        raise argparse.ArgumentError(f"Address {val} is not valid")


def strip_option(option: str):
    return option.replace("raw_binary_", "")


class RawBinary:
    def __init__(self,
                 file_format: str = "unknown",
                 cpu_architecture: Optional[str] = "x86-64",
                 function_entry_point_detection_needed: Optional[bool] = True,
                 image_base_address: Optional[int] = None,
                 function_file_offset: Optional[int] = None):
        self.file_format = file_format
        self.cpu_architecture = cpu_architecture
        self.function_entry_point_detection_needed = function_entry_point_detection_needed
        self.image_base_address = image_base_address
        self.function_file_offset = function_file_offset

    OPTIONS = ['raw_binary_file_format', 'raw_binary_cpu_architecture', 'raw_binary_image_base_address',
               'raw_binary_function_entry_point_detection_needed', 'raw_binary_function_file_offset']

    @property
    def file_format(self) -> str:
        return self.__file_format

    @file_format.setter
    def file_format(self, file_format: str):
        self.__file_format = file_format

    @property
    def cpu_architecture(self) -> str:
        return self.__cpu_architecture

    @cpu_architecture.setter
    def cpu_architecture(self, cpu_architecture: str):
        self.__cpu_architecture = cpu_architecture

    @property
    def function_entry_point_detection_needed(self) -> bool:
        return self.__function_entry_point_detection_needed

    @function_entry_point_detection_needed.setter
    def function_entry_point_detection_needed(self,
                                              function_entry_point_detection_needed: Optional[bool] = True) -> None:
        self.__function_entry_point_detection_needed = function_entry_point_detection_needed

    @property
    def image_base_address(self) -> Optional[int]:
        return self.__image_base_address

    @image_base_address.setter
    def image_base_address(self, image_base_address):
        self.__image_base_address = image_base_address

    @property
    def function_file_offset(self) -> Optional[int]:
        return self.__function_file_offset

    @function_file_offset.setter
    def function_file_offset(self, function_file_offset):
        self.__function_file_offset = function_file_offset

    @classmethod
    def create(cls, args):
        raw_binary = None
        for option in RawBinary.OPTIONS:
            if hasattr(args, option) and getattr(args, option) is not None:
                if not raw_binary:
                    raw_binary = RawBinary()
                setattr(raw_binary, strip_option(option), getattr(args, option))
        return raw_binary

    def to_dict(self) -> Dict:
        raw_binary = {}
        for option in RawBinary.OPTIONS:
            raw_binary[option] = getattr(self, strip_option(option))
        return raw_binary


def post_file(label, analysis_mode, first_seen, timeout, environments, disable_network, priority, file_path,
              raw_binary: Optional['RawBinary'] = None, wait=False):
    data = {
        'label': label,
        'analysis_mode': analysis_mode,
        'first_seen': int(first_seen),
        'timeout': timeout,
        'environments': environments,
        'enable_network': not disable_network,
        'priority': priority
    }

    if raw_binary:
        data = {**data, **raw_binary.to_dict()}

    files = {'file': open(file_path, 'rb')}

    resp = requests.post(f'{api.get_api_url()}/v1/submissions/samples', data=data, files=files,
                         headers=api.get_authorization_header())
    if resp.status_code != 201:
        try:
            msg = resp.json()["error"]["message"]
        except:
            msg = ''

        print(f'Failed to submit file for analysis: {msg if msg else resp.text}')
        return

    data = resp.json()
    if data['submissions']:
        print(f'Submitted file for analysis, created {len(data["submissions"])} submission(s).')
    else:
        print('No submissions were created for the submitted file, the file(s) might be unsupported.')
        return

    if wait:
        print('Waiting for analysis to complete.')

        is_done = [False] * len(data['submissions'])
        while True:
            time.sleep(5)
            for idx, submission in enumerate(data['submissions']):
                resp = requests.get(f'{api.get_api_url()}/v1/tasks/{submission["task_id"]}',
                                    headers=api.get_authorization_header())
                if resp.status_code != 200:
                    continue

                submission_data = resp.json()
                if submission_data['status'] not in ('queued', 'analyzing'):
                    is_done[idx] = True

            if all(is_done):
                break

        for idx, submission in enumerate(data['submissions']):
            print(f'\nReport for analysis #{idx + 1}:')
            if submission['status'] not in ('unsupported', 'failed'):
                get_sample_by_analysis(submission['analysis']['id'])
            else:
                print(f'Analysis status: {submission["status"]}')


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Submit a file for analysis')
    parser.add_argument('--file-path', required=True, type=str,
                        help='Path to the file, can be a zip file.')
    parser.add_argument('--analysis-mode', required=False, type=str, default='dynamic',
                        help='Processing mode, can be: dynamic, static')
    parser.add_argument('--raw-binary-file-format', required=False, choices=['unknown', 'raw', 'pe'],
                        help='Raw binary file format')
    parser.add_argument('--raw-binary-cpu-architecture', required=False, choices=['x86-64', 'x86-32'],
                        help='Raw binary CPU architecture')
    parser.add_argument('--raw-binary-image-base-address', required=False,
                        help='Raw binary image base address, accepted are decimal and hexadecimal numbers')
    parser.add_argument('--raw-binary-function-entry-point-detection-needed', required=False, type=bool,
                        help='Raw binary function entry point detection needed')
    parser.add_argument('--raw-binary-function-file-offset', required=False, type=memory_address,
                        help='Raw binary function file offset, accepted are decimal and hexadecimal numbers')
    parser.add_argument('--label', required=False, type=str, default='',
                        help='Label of the file.')
    parser.add_argument('--disable-network', required=False, default=False, action='store_true',
                        help='Disable network access in the VM.')
    parser.add_argument('--first-seen', required=False, type=int, default=time.mktime(datetime.now().timetuple()),
                        help='Unix timestamp of first seen date.')
    parser.add_argument('--timeout', required=False, type=int, default=180,
                        help='Amount of seconds the sample is allowed to run.')
    parser.add_argument('--environments', required=False, nargs='+',
                        help='List of sandbox VMs to use: win10_1709_x64, win7_sp1_x64, win7_sp1_x86')
    parser.add_argument('--priority', required=False, default=1, type=int, choices=range(0, 10),
                        help='Priority (0=low, 9=high).')
    parser.add_argument('--wait', required=False, default=False, action='store_true',
                        help='Wait for the analysis to complete and show a report afterwards.')

    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()

    raw_binary = None
    if args.analysis_mode == 'static':
        raw_binary = RawBinary.create(args)

    post_file(args.label, args.analysis_mode, args.first_seen, args.timeout, args.environments, args.disable_network,
              args.priority, args.file_path, raw_binary, args.wait)
