import argparse
import glob
import os
import sys
import textwrap
import time
from datetime import datetime
from typing import Optional, Dict, List

import requests
from threatray.api import API

api = API()


def post_file(label, analysis_mode, first_seen, timeout, environments, disable_network, priority, file_path,
              raw_binary: Optional['RawBinary']) -> Optional[List['Analysis']]:
    data = {
        'label': label,
        'analysis_mode': analysis_mode,
        'first_seen': int(first_seen),
        'timeout': timeout,
        'environments': environments,
        'enable_network': not disable_network,
        'priority': priority
    }

    if raw_binary:
        data = {**data, **raw_binary.to_dict()}

    files = {'file': open(file_path, 'rb')}
    resp = requests.post(f'{api.get_api_url()}/v1/submissions/samples', data=data, files=files,
                         headers=api.get_authorization_header())
    if resp.status_code != 201:
        try:
            msg = resp.json()["error"]["message"]
        except:
            msg = ''

        print(f'INFO: Failed to submit file for analysis: {msg if msg else resp.text}')
        return

    return Analysis.from_submissions(resp.json()['submissions'])


def strip_option(option: str):
    return option.replace("raw_binary_", "")


class RawBinary:
    def __init__(self,
                 file_format: str = "unknown",
                 cpu_architecture: Optional[str] = "x86-64",
                 function_entry_point_detection_needed: Optional[bool] = True,
                 image_base_address: Optional[int] = None,
                 function_file_offset: Optional[int] = None):
        self.file_format = file_format
        self.cpu_architecture = cpu_architecture
        self.function_entry_point_detection_needed = function_entry_point_detection_needed
        self.image_base_address = image_base_address
        self.function_file_offset = function_file_offset

    OPTIONS = ['raw_binary_file_format', 'raw_binary_cpu_architecture', 'raw_binary_image_base_address',
               'raw_binary_function_entry_point_detection_needed', 'raw_binary_function_file_offset']

    @property
    def file_format(self) -> str:
        return self.__file_format

    @file_format.setter
    def file_format(self, file_format: str):
        self.__file_format = file_format

    @property
    def cpu_architecture(self) -> str:
        return self.__cpu_architecture

    @cpu_architecture.setter
    def cpu_architecture(self, cpu_architecture: str):
        self.__cpu_architecture = cpu_architecture

    @property
    def function_entry_point_detection_needed(self) -> bool:
        return self.__function_entry_point_detection_needed

    @function_entry_point_detection_needed.setter
    def function_entry_point_detection_needed(self,
                                              function_entry_point_detection_needed: Optional[bool] = True) -> None:
        self.__function_entry_point_detection_needed = function_entry_point_detection_needed

    @property
    def image_base_address(self) -> Optional[int]:
        return self.__image_base_address

    @image_base_address.setter
    def image_base_address(self, image_base_address):
        self.__image_base_address = image_base_address

    @property
    def function_file_offset(self) -> Optional[int]:
        return self.__function_file_offset

    @function_file_offset.setter
    def function_file_offset(self, function_file_offset):
        self.__function_file_offset = function_file_offset

    @classmethod
    def create(cls, args):
        raw_binary = None
        for option in RawBinary.OPTIONS:
            if hasattr(args, option) and getattr(args, option) is not None:
                if not raw_binary:
                    raw_binary = RawBinary()
                setattr(raw_binary, strip_option(option), getattr(args, option))
        return raw_binary

    def to_dict(self) -> Dict:
        raw_binary = {}
        for option in RawBinary.OPTIONS:
            raw_binary[option] = getattr(self, strip_option(option))
        return raw_binary


class Analysis:
    def __init__(self, task_id: int, analysis_id: str, file_name: str, sha256hash: str, threats: List[str] = None,
                 analysis_status: str = None):
        if threats is None:
            threats = []
        self.task_id = task_id
        self.id = analysis_id
        self.file_name = file_name
        self.sha256hash = sha256hash
        self.threats = threats
        self.analysis_status = analysis_status

    @property
    def id(self) -> str:
        return self._id

    @property
    def task_id(self) -> str:
        return self._task_id

    @property
    def file_name(self) -> str:
        return self._file_name

    @property
    def sha256hash(self) -> str:
        return self._sha256hash

    @property
    def threats(self) -> List[str]:
        return self._threats

    @property
    def analysis_status(self) -> str:
        return self._analysis_status

    @task_id.setter
    def task_id(self, value):
        self._task_id = value

    @id.setter
    def id(self, value):
        self._id = value

    @file_name.setter
    def file_name(self, value: str):
        self._file_name = value

    @sha256hash.setter
    def sha256hash(self, value: str):
        self._sha256hash = value

    @threats.setter
    def threats(self, value: List[str]):
        self._threats = value

    @analysis_status.setter
    def analysis_status(self, value: Optional[str]):
        self._analysis_status = value

    @staticmethod
    def from_submissions(submissions: List[Dict]) -> List['Analysis']:
        """
        [{'extracted_from': None, 'priority': 1, 'created_at': 1628789710,
        'modified_at': 1628789710, 'submission_id': 'e8b32bcf-3bde-4181-9e35-eddc45b55667',
        'status': 'analyzing', 'username': 'user', 'execution_report': [],
        'sample': {'file_name': '211a56fe55fe7ecc4f65690e82085e28692ce1fbe882bc54384c322813b20cb23.exe',
        'label': 'tf_tests', 'hash_md5': 'e55d2b18d697d5d7ef8ec1a65f9f41a7',
        'hash_sha1': 'cc5470a8634b2905c9a54056152af72de9cfde4d',
        'hash_sha256': '324947c4bc77ab58d64e0fa446f06571c41061de2551386145458672f90845fc',
        'file_type': 'Exe (PE, x86-32)',
        'first_seen': 1628672109, 'threats': [], 'verdict': 'unknown', 'id': 6008},
        'analysis': {'id': 'aade9483-19b7-4eb8-8319-af4b5488e784', 'analysis_mode': 'static',
        'timeout': 180, 'environment': None, 'enable_network': True,
        'threats': [], 'verdict': 'unknown'}, 'task_id': 7163}], 'error': None}]
        """
        analysis_results = []
        for submission in submissions:
            sample = submission.get('sample', {})
            analysis = submission.get('analysis', {})
            status = submission.get('status', None)
            analysis_results.append(Analysis(task_id=submission.get('task_id', None),
                                             sha256hash=sample.get('hash_sha256', None),
                                             file_name=sample.get('file_name', None),
                                             analysis_id=analysis.get('id', None),
                                             analysis_status=status))
        return analysis_results


def submit_batch(args, raw_binary, file_paths) -> List[Analysis]:
    responses: List[Analysis] = []
    for file_path in file_paths:
        responses.extend(
            post_file(args.label, args.analysis_mode, args.first_seen, args.timeout, args.environments,
                      args.disable_network, args.priority, file_path, raw_binary))
    return responses


def update_analysis_status(analysis) -> None:
    req_path = f'{api.get_api_url()}/v1/tasks/{analysis.task_id}'
    resp = requests.get(req_path, headers=api.get_authorization_header())
    if resp.status_code != 200:
        print(f"INFO: invalid status code for {req_path}, {resp.status_code}")
        return None

    submission_data = resp.json()
    analysis.analysis_status = submission_data['status']


def fetch_threats_and_verdicts(analyses):
    """{"sample": {"label": "tf_tests", "hash_md5": "e55d2b18d697d5d7ef8ec1a65f9f41a7",
    "hash_sha1": "cc5470a8634b2905c9a54056152af72de9cfde4d",
    "hash_sha256": "324947c4bc77ab58d64e0fa446f06571c41061de2551386145458672f90845fc",
     "first_seen": 1628672109,
     "magic": "PE32 executable (console) Intel 80386 Mono/.Net assembly, for MS Windows",
     "file_type": "Exe (PE, x86-32)", "file_size": 255931,
     "file_name": "211a56fe55fe7ecc4f65690e82085e28692ce1fbe882bc54384c322813b20cb23.exe",
     "threats": [], "verdict": "unknown", "is_pe": true,
     "downloaded_from": [], "osint": [],
     "static_analysis": {"function_count": 3, "signature_id": null, "signatures": [],
            "threats": [], "verdict": "malicious"}, "scope": "private"},
    "analysis": {"id": "44d59766-b844-457b-be42-472cc82e916e", "creation_time": 1628793069, "environment": null,
    "execution_report": [], "type": "static", "threats": [],
    "verdict": "unknown", "scope": "private"},
    "processes": null, "ioc": null}
    """
    for analysis in analyses:
        req_path = f'{api.get_api_url()}/v1/samples/{analysis.sha256hash}/analyses/{analysis.id}'
        resp = requests.get(req_path, headers=api.get_authorization_header())
        if resp.status_code != 200:
            print(f"INFO: invalid status code for {req_path}, {resp.status_code}")
            return None

        response = resp.json()
        analysis_result = response.get('analysis')
        analysis.verdict = analysis_result['verdict']
        analysis.threats = analysis_result['threats']


def analysis_output(analyses) -> str:
    output = "file_name,sha256hash,threats,verdicts\n"

    for analysis in analyses:
        threats = ', '.join(
            map(lambda threat: f"{threat.get('label')} ({threat.get('confidence')})", analysis.threats))
        output += f"{analysis.file_name},{analysis.sha256hash},{threats},{analysis.verdict}\n"
    return output


def batch_process(args, raw_binary, file_paths: List) -> List[Analysis]:
    analyses: List[Analysis] = submit_batch(args, raw_binary, file_paths)
    analyses_finished = False
    unfinished_analyses = analyses.copy()
    while not analyses_finished:
        for analysis in unfinished_analyses:
            update_analysis_status(analysis)
        analyses_finished = not any(
            map(lambda a: a.analysis_status not in ["done", "unsupported"], unfinished_analyses))
        unfinished_analyses = [a for a in unfinished_analyses if a.analysis_status not in ["done", "unsupported"]]
        print(f"INFO: {len(unfinished_analyses)} unfinished analyses in batch")
        time.sleep(2)
    fetch_threats_and_verdicts(analyses)
    return analyses


def get_file_names(folder_path):
    return glob.glob(os.path.join(folder_path, "*"))


def path(file_path):
    if not os.path.exists(file_path):
        raise ValueError(f"Path {file_path} does not exist")
    return file_path


def file_path_type(file_path):
    directory_path = os.path.exists(os.path.dirname(os.path.abspath(file_path)))
    if not directory_path:
        raise ValueError(f"Directory {directory_path} does not exist")

    if os.path.exists(file_path):
        raise ValueError(f"File {file_path} already exists")
    return file_path


def memory_address(val) -> int:
    try:
        if isinstance(val, int):
            return int(val)
        if isinstance(val, str):
            return int(val, 16)
    except ValueError as _:
        raise argparse.ArgumentError(f"Address {val} is not valid")


def main():
    parser = __setup_parser()
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)
    args = parser.parse_args()

    raw_binary = None
    if args.analysis_mode == 'static':
        raw_binary = RawBinary.create(args)
    files_in_directory = get_file_names(args.src_directory)
    output = __process_batch_request(args, files_in_directory, raw_binary)
    print(output)
    if args.output_file:
        with open(args.output_file, "w") as f:
            f.write(output)


def __process_batch_request(args, files_in_directory, raw_binary):
    batch_size = 100
    batches = [files_in_directory[i:i + batch_size] for i in range(0, len(files_in_directory), batch_size)]
    analyses: List[Analysis] = []
    for batch in batches:
        print("INFO: Starting with batch")
        analyses.extend(batch_process(args, raw_binary, batch))
    output = analysis_output(analyses)
    return output


def __setup_parser():
    epilog = """Submit and get the threats and verdicts for all samples which are located in a folder.
The submission config will be the same for all submissions.

EXAMPLE:

python3 batch_process_and_analyze.py --src-directory /home/user/Downloads/binaries \
--analysis-mode static --label binaries_from_case --raw-binary-file-format \
unknown --output-file /tmp/output-analysis.csv"""

    parser = argparse.ArgumentParser(
        description='Submit a directory for analysis and get the verdicts, may scarlett be with you',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent(epilog))
    parser.add_argument('--src-directory', required=True, type=path,
                        help="Path to a directory where all samples are located in you'd like to submit in the batch")
    parser.add_argument('--output-file', required=False, type=file_path_type,
                        help="Destination file the output is written to")
    parser.add_argument('--analysis-mode', required=False, type=str, default='dynamic',
                        help='Processing mode, can be: dynamic, static')
    parser.add_argument('--label', required=False, type=str, default='', help='Label of the file.')
    parser.add_argument('--disable-network', required=False, default=False, action='store_true',
                        help='Disable network access in the VM.')
    parser.add_argument('--first-seen', required=False, type=int, default=time.mktime(datetime.now().timetuple()),
                        help='Unix timestamp of first seen date.')
    parser.add_argument('--timeout', required=False, type=int, default=180,
                        help='Amount of seconds the sample is allowed to run.')
    parser.add_argument('--environments', required=False, nargs='+',
                        help='List of sandbox VMs to use: win10_1709_x64, win7_sp1_x64, win7_sp1_x86')
    parser.add_argument('--raw-binary-file-format', required=False, choices=['unknown', 'raw', 'pe'],
                        help='Raw binary file format')
    parser.add_argument('--raw-binary-cpu-architecture', required=False, choices=['x86-64', 'x86-32'],
                        help='Raw binary CPU architecture')
    parser.add_argument('--raw-binary-image-base-address', required=False,
                        help='Raw binary image base address, accepted are decimal and hexadecimal numbers')
    parser.add_argument('--raw-binary-function-entry-point-detection-needed', required=False, type=bool,
                        help='Raw binary function entry point detection needed')
    parser.add_argument('--raw-binary-function-file-offset', required=False, type=memory_address,
                        help='Raw binary function file offset, accepted are decimal and hexadecimal numbers')
    parser.add_argument('--priority', required=False, default=1, type=int, choices=range(0, 10),
                        help='Priority (0=low, 9=high).')
    return parser


if __name__ == "__main__":
    main()
