import argparse
import datetime
import shutil
import sys
from typing import Optional

import requests
from threatray.api import API

api = API()


def post_file(label: str, file_path: str):
    data = {
        'label': label,
    }

    files = {'archive': open(file_path, 'rb')}

    resp = requests.post(f'{api.get_api_url()}/v1/submissions/endpoint-scan-archive',
                         data=data,
                         files=files,
                         headers=api.get_authorization_header())
    if resp.status_code != 201:
        try:
            msg = resp.json()["error"]["message"]
        except:
            msg = ''

        print(f'Failed to submit archive for analysis: {msg if msg else resp.text}')
        return

    data = resp.json()
    if data['submissions']:
        print(f'Submitted file for analysis, created {len(data["submissions"])} submission(s).')
    else:
        print('No submissions were created for the submitted archive.')
        return


def create_binary(expires_at: int):
    data = {
        'expires_at': expires_at
    }
    resp = requests.post(f'{api.get_api_url()}/v1/binary-authentication/binaries',
                         json=data,
                         headers=api.get_authorization_header())

    if resp.status_code != 200:
        print(f'Failed to create new binary: {resp} {resp.text}')
        return

    data = resp.json()
    print(f"Created new binary with hash: {data['sha256hash']}")


def format_date_from_optional_timestamp(optional_timestamp: Optional[int]) -> str:
    if optional_timestamp is None:
        return ''
    else:
        return str(datetime.datetime.fromtimestamp(optional_timestamp))


def list_binaries():
    resp = requests.get(f'{api.get_api_url()}/v1/binary-authentication/binaries',
                        headers=api.get_authorization_header())

    if resp.status_code != 200:
        print(f'Failed to create new binary: {resp}')
        return

    data = resp.json()

    print('{:<64} {:<20} {:<20}'.format('Binary SHA-256', "Expires at", 'Deactivated at'))
    for binary in data['binaries']:
        print('{:<64} {:<20} {:<20}'.format(binary["sha256hash"],
                                            format_date_from_optional_timestamp(binary["expires_at"]),
                                            format_date_from_optional_timestamp(binary["deactivated_at"])))


def download(sha256sum: str):
    file = 'threatray_endpoint_scan.exe'
    url = f'{api.get_api_url()}/v1/binary-authentication/binaries/{sha256sum}/download'
    with requests.get(url, stream=True, headers=api.get_authorization_header()) as r:
        if r.status_code == 404:
            print(f"Binary with hash {sha256sum} not found")
            return
        elif r.status_code != 200:
            print(f"Failed to fetch binary {r}")
            return
        with open(file, 'wb') as f:
            shutil.copyfileobj(r.raw, f)

    print(f"Binary with hash {sha256sum} was downloaded to {file}")


def download_ps1(sha256sum: str):
    file = 'threatray_endpoint_scan.ps1'
    url = f'{api.get_api_url()}/v1/binary-authentication/binaries/{sha256sum}/ps1'
    with requests.get(url, stream=True, headers=api.get_authorization_header()) as r:
        if r.status_code == 404:
            print(f"Binary with hash {sha256sum} not found")
            return
        elif r.status_code != 200:
            print(f"Failed to fetch powershell script for binary {r}")
            return
        with open(file, 'wb') as f:
            shutil.copyfileobj(r.raw, f)

    print(f"Powershell script for binary with hash {sha256sum} was downloaded to {file}")


def deactivate(sha256sum: str):
    resp = requests.post(f'{api.get_api_url()}/v1/binary-authentication/binaries/{sha256sum}/deactivate',
                         headers=api.get_authorization_header())

    if resp.status_code == 404:
        print(f"Binary with hash {sha256sum} not found")
        return
    elif resp.status_code != 200:
        print(f'Failed to deactivate the binary: {resp}')
        return

    print(f"Binary with hash {sha256sum} has been deactivated.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Endpoint scan manager')
    subparsers = parser.add_subparsers(dest='mode')

    upload_parser = subparsers.add_parser("upload")
    upload_parser.add_argument('--file-path', required=True, type=str,
                               help='Path to the endpoint scan archive file.')
    upload_parser.add_argument('--label', required=False, type=str, default='',
                               help='Label of the analysis.')

    create_binary_parser = subparsers.add_parser('create-binary')
    create_binary_parser.add_argument('--expires-in-days', required=False, type=int,
                                      default=356, help="Days until the binary expires.")

    subparsers.add_parser('list-binaries')

    download_binary_parser = subparsers.add_parser('download-binary')
    download_binary_parser.add_argument('--sha256', required=True, type=str,
                                        help='SHA-256 hash of a binary to download.')

    download_ps1_parser = subparsers.add_parser('download-ps1')
    download_ps1_parser.add_argument('--sha256', required=True, type=str,
                                     help='SHA-256 hash of a binary to download as a Powershell script.')

    deactivate_binary_parser = subparsers.add_parser("deactivate-binary")
    deactivate_binary_parser.add_argument('--sha256', required=True, type=str,
                                          help='SHA-256 hash of the binary to deactivate.')

    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()

    if args.mode == 'upload':
        post_file(args.label, args.file_path)
    elif args.mode == 'create-binary':
        expires_at = int((datetime.datetime.now() + datetime.timedelta(days=args.expires_in_days)).timestamp())
        create_binary(expires_at)
    elif args.mode == 'list-binaries':
        list_binaries()
    elif args.mode == 'download-binary':
        download(args.sha256)
    elif args.mode == 'download-ps1':
        download_ps1(args.sha256)
    elif args.mode == 'deactivate-binary':
        deactivate(args.sha256)
