import argparse
import sys

import requests
from threatray.api import API

api = API()


def sanitize_value(value):
    # Remove any starting or ending " and escape the remainder of them
    if value[0] == '"':
        value = value[1:]
    if value[-1] == '"':
        value = value[0:len(value) - 1]
    value.replace('"', '\"')
    return value


def search(selector, value, scope='both'):
    value = sanitize_value(value)

    query = f'{selector}: "{value}"'
    print(f'Query: {query}')
    resp = requests.get(f'{api.get_api_url()}/v1/search', params={'query': query, 'scope': scope},
                        headers=api.get_authorization_header())
    if resp.status_code != 200:
        print(f'Failed to search: {resp.json()["error"]}')
        return

    data = resp.json()
    if 'analyses' not in data or 'aggregations' not in data:
        print('No results found.')
        return

    print(f'Found {len(data["analyses"])} matching analyses.')
    print('Threats:')
    if 'aggregations' in data and 'threats' in data['aggregations']:
        for threat in data["aggregations"]["threats"]:
            print(f'  {threat["key"]}: {threat["count"]}')


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Search for samples by IOCs')
    parser.add_argument('--domain', required=False, type=str, help='Search by domain IOCs.')
    parser.add_argument('--ip', required=False, type=str, help='Search by IP IOCs.')
    parser.add_argument('--url', required=False, type=str, help='Search by URL IOCs.')
    parser.add_argument('--mutex', required=False, type=str, help='Search by mutex IOCs.')
    parser.add_argument('--registry', required=False, type=str, help='Search by registry IOCs.')
    parser.add_argument('--file', required=False, type=str, help='Search by file IOCs.')
    parser.add_argument('--retrohunt', required=False, type=str, help='Retrohunt by sample hash or analysis ID.')
    parser.add_argument('--retrohunt-memory', required=False, type=str, help='Retrohunt by memory region hash.')
    parser.add_argument('--process', required=False, type=str, help='Search by process and command line.')
    parser.add_argument('--signature', required=False, type=str, help='Search by threat signature name.')
    parser.add_argument('--label', required=False, type=str, help='Search by analysis label.')
    parser.add_argument('--scope', required=False, type=str, default='both', choices=['both', 'private', 'public'],
                        help='Limit search to a scope.')

    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()

    selector = None
    if args.domain:
        selector = 'domain'

    if args.ip:
        selector = 'ip'

    if args.url:
        selector = 'url'

    if args.mutex:
        selector = 'mutex'

    if args.registry:
        selector = 'registry'

    if args.file:
        selector = 'file'

    if args.retrohunt:
        selector = 'retrohunt'

    if args.retrohunt_memory:
        selector = 'retrohunt-memory'

    if args.process:
        selector = 'process'

    if args.signature:
        selector = 'signature'

    if args.label:
        selector = 'label'

    if not selector:
        print('No selector given, exiting...')
        sys.exit(1)

    ret = search(selector, vars(args).get(selector), args.scope)
