import argparse
import sys
from datetime import datetime

import requests
from threatray.api import API

api = API()


def get_sample_by_hash(sample_hash, analysis_id=None):
    if analysis_id is None:
        res = requests.get(f'{api.get_api_url()}/v1/samples/{sample_hash}', headers=api.get_authorization_header())
    else:
        res = requests.get(f'{api.get_api_url()}/v1/samples/{sample_hash}/analyses/{analysis_id}',
                           headers=api.get_authorization_header())

    if res.status_code == 200:
        report = res.json()
        show_report(report)
    else:
        print(f'Sample with hash {sample_hash} was not found.')


def get_sample_by_analysis(analysis_id):
    res = requests.get(f'{api.get_api_url()}/v1/analyses/{analysis_id}',
                       headers=api.get_authorization_header())

    if res.status_code == 200:
        report = res.json()
        show_report(report)
    else:
        print(f'Analysis with ID {analysis_id} was not found.')


def show_report(report):
    print(f'File name: {report["sample"]["file_name"]}')
    print(f'File type: {report["sample"]["file_type"]}')
    print(f'MD5: {report["sample"]["hash_md5"]}')
    print(f'SHA1: {report["sample"]["hash_sha1"]}')
    print(f'SHA256: {report["sample"]["hash_sha256"]}')
    print(f'Verdict: {report["analysis"]["verdict"]}')
    print(f'Threats: {[t["label"] for t in report["analysis"]["threats"]]}')
    print(f'Analysis creation time: {datetime.fromtimestamp(report["analysis"]["creation_time"])}')

    if report["processes"]:
        print('Processes:')
        for process in report["processes"]:
            print(f'  Process #{process["process_id"]}: {process["name"]} [{process["pid"]}], '
                  f'Threats: {[t["label"] for t in process["threats"]]}')

    if report["ioc"]:
        print('IOCs:')
        print(f'  Domains: {len(report["ioc"]["domains"])}')
        print(f'  URLs: {len(report["ioc"]["urls"])}')
        print(f'  IPs: {len(report["ioc"]["ips"])}')
        print(f'  Files: {len(report["ioc"]["files"])}')
        print(f'  Mutexes: {len(report["ioc"]["mutexes"])}')
        print(f'  Registry: {len(report["ioc"]["registry"])}')


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Show the report of a sample.')
    parser.add_argument('--hash', required=False, type=str, help='Hash (MD5, SHA1, SHA256) of the file to show.')
    parser.add_argument('--analysis', required=False, type=str, help='Analysis UUID of the file to show.')

    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()

    if not (args.hash or args.analysis):
        print('Need to give at least --hash or --analyses, exiting...')
        sys.exit(1)

    if args.hash:
        get_sample_by_hash(args.hash, args.analysis if args.analysis else None)
    else:
        get_sample_by_analysis(args.analysis)
